import os

import trio
import pytest_trio

from trio_bybit.streams import BybitSocketManager


async def pub_stream_check(socket):
    async for msg in socket.get_next_message():
        assert "ts" in msg
        assert "type" in msg
        assert "data" in msg


async def pub_unsub(socket):
    unsub = {
        "op": "unsubscribe",
        "args": ["orderbook.1.BTCUSDT", "publicTrade.BTCUSDT"],
    }
    await socket.subscribe(unsub)


async def test_public_stream():
    socket = BybitSocketManager()
    with trio.move_on_after(12):
        async with socket.connect():
            subscription = {
                "op": "subscribe",
                "args": ["orderbook.1.BTCUSDT", "publicTrade.BTCUSDT"],
            }
            await socket.subscribe(subscription)
            async with trio.open_nursery() as nursery:
                nursery.start_soon(pub_stream_check, socket)
                await trio.sleep(10)
                nursery.start_soon(pub_unsub, socket)


async def test_public_linear_stream():
    socket = BybitSocketManager(endpoint="linear")
    with trio.move_on_after(30):
        async with socket.connect():
            subscription = {
                "op": "subscribe",
                "args": ["orderbook.1.BTCUSDT", "publicTrade.BTCUSDT"],
            }
            await socket.subscribe(subscription)

            async for msg in socket.get_next_message():
                if msg.get("topic") == "orderbook.1.BTCUSDT":
                    assert "topic" in msg
                    assert "type" in msg
                    assert "data" in msg
                    assert "s" in msg["data"]
                    assert "a" in msg["data"]
                    assert "b" in msg["data"]
                elif msg.get("topic") == "publicTrade.BTCUSDT":
                    assert "topic" in msg
                    assert "type" in msg
                    assert "data" in msg
                    assert "S" in msg["data"][0]
                    assert "v" in msg["data"][0]
                    assert "p" in msg["data"][0]


async def test_private_stream():
    socket = BybitSocketManager(
        endpoint="private",
        api_key=os.getenv("BYBIT_API_KEY"),
        api_secret=os.getenv("BYBIT_API_SECRET"),
    )
    with trio.move_on_after(100):
        async with socket.connect():
            subscription = {
                "op": "subscribe",
                "args": ["order"],
            }
            await socket.subscribe(subscription)

            async for msg in socket.get_next_message():
                print(msg)


async def test_private_stream_by_rsa_sig():
    socket = BybitSocketManager(
        endpoint="private",
        api_key=os.getenv("BYBIT_RSA_API_KEY"),
        api_secret=os.getenv("BYBIT_RSA_PRIVATE_KEY_PATH"),
        sign_style="RSA",
    )
    with trio.move_on_after(100):
        async with socket.connect():
            subscription = {
                "op": "subscribe",
                "args": ["order"],
            }
            await socket.subscribe(subscription)

            async for msg in socket.get_next_message():
                print(msg)



import base64
import hmac
import os

import httpx
import hashlib
import time

import orjson
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.serialization import load_pem_private_key

from .exceptions import BybitAPIException, BybitRequestException


class BaseClient:
    API_URL = "https://api.bybit.com/"
    # SECONDARY_API_URL = "https://api.bytick.com/"
    TEST_NET_API_URL = "https://api-testnet.bybit.com/"
    DEMO_NET_API_URL = "https://api-demo.bybit.com/"
    API_VERSION = "v5"

    REQUEST_TIMEOUT: float = 5

    def __init__(
        self,
        api_key: str | None = None,
        api_secret: str | None = None,
        receive_window: int = 5000,
        sign_style: str = "HMAC",
        api_secret_passphrase: bytes | None = None,
        alternative_net: str = "",
    ):
        """API Client constructor
        :param api_key: Api Key
        :param api_secret: Api Secret. For RSA sign style, it should be the path to the private key file
        :param receive_window: Receive Window
        :param sign_style: Sign Style. Default HMAC. Choices: ["HMAC", "RSA"]
        :param api_secret_passphrase: RSA (or other cryptography method) private key passphrase
        :param alternative_net: Alternative network. Default empty to use mainnet. Choices: "test", "demo"
        """
        self.API_KEY = api_key
        self.API_SECRET = api_secret
        self.API_SECRET: RSAPrivateKey | str
        self.sign_style = sign_style
        if self.sign_style != "HMAC":
            with open(api_secret, "rb") as f:
                self.API_SECRET = load_pem_private_key(f.read(), password=api_secret_passphrase)
        else:
            self.API_SECRET = api_secret
        self.receive_window = receive_window
        if sign_style != "HMAC" and sign_style != "RSA":
            raise ValueError("Invalid sign style. Must be HMAC or RSA")
        self.sign_style = sign_style
        self.timestamp_offset = 0
        if alternative_net == "test":
            self.base = self.TEST_NET_API_URL
        elif alternative_net == "demo":
            self.base = self.DEMO_NET_API_URL
        else:
            self.base = self.API_URL
        self.base_url = self.base + self.API_VERSION + "/"
        self.session = httpx.AsyncClient(http2=True, base_url=self.base_url)

    def _get_headers(self, timestamp_milli: int, signed=False, timeout: int = None) -> dict:
        headers = {
            "X-BAPI-TIMESTAMP": str(timestamp_milli),
            "X-BAPI-RECV-WINDOW": str(self.receive_window),
        }
        if signed:
            headers["X-BAPI-API-KEY"] = self.API_KEY
        return headers

    def _create_api_uri(self, path: str) -> httpx.URL:
        return httpx.URL(os.path.join(self.base, self.API_VERSION, path))

    def _generate_signature(self, request: httpx.Request, timestamp_milli: int) -> str:
        if request.method == "GET":
            prepared_str = str(timestamp_milli) + self.API_KEY + str(self.receive_window) + str(request.url.params)
        else:
            prepared_str = (
                str(timestamp_milli) + self.API_KEY + str(self.receive_window) + request.content.decode("utf-8")
            )
        prepared_bytes = prepared_str.encode("utf-8")
        if self.sign_style == "HMAC":
            return hmac.new(self.API_SECRET.encode("utf-8"), prepared_bytes, hashlib.sha256).hexdigest()
        else:  # RSA
            signature = self.API_SECRET.sign(prepared_bytes, padding.PKCS1v15(), hashes.SHA256())
            return base64.b64encode(signature).decode()

    def _get_request(self, method, uri, signed: bool, **kwargs) -> httpx.Request:
        timestamp = int(time.time() * 1000 + self.timestamp_offset)
        headers = self._get_headers(timestamp, signed)
        if method.lower() == "get":
            req = self.session.build_request(method, uri, headers=headers, params=kwargs)
        else:
            req = self.session.build_request(method, uri, headers=headers, json=kwargs)
        if signed:
            req.headers["X-BAPI-SIGN"] = self._generate_signature(req, timestamp)
            req.headers["X-BAPI-SIGN-TYPE"] = "2"
        return req


class AsyncClient(BaseClient):
    def __init__(
        self,
        api_key: str | None = None,
        api_secret: str | None = None,
        receive_window: int = 5000,
        sign_style: str = "HMAC",
        api_secret_passphrase: bytes | None = None,
        alternative_net: str = "",
    ):
        super().__init__(api_key, api_secret, receive_window, sign_style, api_secret_passphrase, alternative_net)
        self.session: httpx.AsyncClient = httpx.AsyncClient(http2=True)

    @classmethod
    async def create(
        cls,
        api_key: str | None = None,
        api_secret: str | None = None,
        receive_window: int = 5000,
        sign_style: str = "HMAC",
        api_secret_passphrase: bytes | None = None,
        alternative_net: str = "",
    ) -> "AsyncClient":
        self = cls(api_key, api_secret, receive_window, sign_style, api_secret_passphrase, alternative_net)

        try:
            # calculate timestamp offset between local and coinex server
            res = await self.get_server_time()
            self.timestamp_offset = 1000 * (int(res["result"]["timeSecond"]) - int(time.time()))

            return self
        except Exception:
            # If ping throw an exception, the current self must be cleaned
            # else, we can receive an "asyncio:Unclosed client session"
            await self.close_connection()
            raise

    async def __aenter__(self):
        return self

    async def __aexit__(self, *excinfo):
        await self.session.aclose()

    async def close_connection(self):
        if self.session:
            assert self.session
            await self.session.aclose()

    async def _request(self, method, uri: httpx.URL, signed: bool, **kwargs):
        request: httpx.Request = self._get_request(method, uri, signed, **kwargs)
        response = await self.session.send(request)
        return await self._handle_response(response)

    @staticmethod
    async def _handle_response(response: httpx.Response):
        """Internal helper for handling API responses from the server.
        Raises the appropriate exceptions when necessary; otherwise, returns the
        response.
        """
        if response.is_error:
            raise BybitAPIException(response, response.status_code, response.text)
        try:
            return orjson.loads(response.text)
        except ValueError:
            raise BybitRequestException(f"Invalid Response: {response.text}")

    async def _request_api(self, method, path, signed=False, **kwargs):
        uri = self._create_api_uri(path)
        return await self._request(method, uri, signed, **kwargs)

    async def _get(self, path, signed=False, **kwargs):
        return await self._request_api("get", path, signed, **kwargs)

    async def _post(self, path, signed=False, **kwargs) -> dict:
        return await self._request_api("post", path, signed, **kwargs)

    async def _put(self, path, signed=False, **kwargs) -> dict:
        return await self._request_api("put", path, signed, **kwargs)

    async def _delete(self, path, signed=False, **kwargs) -> dict:
        return await self._request_api("delete", path, signed, **kwargs)

    async def get_server_time(self) -> dict:
        return await self._get("market/time")

    async def get_symbol_info(self, **kwargs) -> dict:
        return await self._get("market/instruments-info", **kwargs)

    async def get_orderbook(self, **kwargs) -> dict:
        return await self._get("market/orderbook", **kwargs)

    async def get_klines(self, **kwargs) -> dict:
        return await self._get("market/kline", **kwargs)

    async def get_funding_history(self, **kwargs) -> dict:
        return await self._get("market/funding/history", **kwargs)

    async def get_ticker(self, **kwargs) -> dict:
        return await self._get("market/tickers", **kwargs)

    async def place_order(self, **kwargs) -> dict:
        return await self._post("order/create", **kwargs, signed=True)

    async def amend_order(self, **kwargs) -> dict:
        return await self._post("order/amend", **kwargs, signed=True)

    async def cancel_order(self, **kwargs) -> dict:
        return await self._post("order/cancel", **kwargs, signed=True)

    async def cancel_all_orders(self, **kwargs) -> dict:
        return await self._post("order/cancel-all", **kwargs, signed=True)

    async def get_order_history(self, **kwargs) -> dict:
        return await self._get("order/history", **kwargs, signed=True)

    async def get_open_orders(self, **kwargs) -> dict:
        return await self._get("order/realtime", **kwargs, signed=True)

    async def get_execution_history(self, **kwargs) -> dict:
        return await self._get("execution/list", **kwargs, signed=True)

    async def get_position_info(self, **kwargs) -> dict:
        return await self._get("position/list", **kwargs, signed=True)

    async def set_leverage(self, **kwargs) -> dict:
        return await self._post("position/set-leverage", **kwargs, signed=True)

    async def get_wallet_balance(self, **kwargs) -> dict:
        return await self._get("account/wallet-balance", **kwargs, signed=True)

    async def set_collateral_switch(self, **kwargs) -> dict:
        return await self._post("account/set-collateral-switch", **kwargs, signed=True)

    async def get_borrow_history(self, **kwargs) -> dict:
        return await self._get("account/borrow-history", **kwargs, signed=True)

    async def get_collateral_info(self, **kwargs) -> dict:
        return await self._get("account/collateral-info", **kwargs, signed=True)

    async def get_fee_rate(self, **kwargs) -> dict:
        return await self._get("account/fee-rate", **kwargs, signed=True)

    async def get_transaction_log(self, **kwargs) -> dict:
        return await self._get("account/transaction-log", **kwargs, signed=True)

    async def get_account_info(self) -> dict:
        return await self._get("account/info", signed=True)

    async def set_hedging_mode(self, **kwargs) -> dict:
        return await self._post("account/set-hedging-mode", **kwargs, signed=True)

    async def get_coin_greeks(self, **kwargs) -> dict:
        return await self._get("asset/coin-greeks", **kwargs, signed=True)

    async def get_spot_margin_trade_data(self, **kwargs) -> dict:
        return await self._get("spot-margin-trade/data", **kwargs)

    async def switch_spot_margin_trade_mode(self, **kwargs) -> dict:
        return await self._post("spot-margin-trade/switch-mode", **kwargs, signed=True)

    async def set_spot_margin_trade_leverage(self, **kwargs) -> dict:
        return await self._post("spot-margin-trade/set-leverage", **kwargs, signed=True)

    async def get_spot_margin_trade_state(self, **kwargs) -> dict:
        return await self._get("spot-margin-trade/state", **kwargs, signed=True)



import os

from trio_bybit.client import AsyncClient

# Generate key pair by 3 steps: generate pairs file, extract pub key and private key
# openssl genrsa -out keypair.pem 2048
# openssl rsa -in keypair.pem -pubout -out publickey.crt
# openssl pkcs8 -topk8 -inform PEM -outform PEM -nocrypt -in keypair.pem -out pkcs8.key


async def test_get_wallet():
    client = await AsyncClient.create(
        api_key=os.getenv("BYBIT_API_KEY"),
        api_secret=os.getenv("BYBIT_PRIVATE_KEY_PATH"),
        sign_style="RSA",
    )

    async with client:
        wallet = await client.get_wallet_balance(accountType="UNIFIED")
        assert wallet["retCode"] == 0
        assert wallet["retMsg"] == "OK"
        assert wallet["result"]["list"][0]["accountType"] == "UNIFIED"
        position = await client.get_position_info(category="linear", settleCoin="USDT")
        assert position["retCode"] == 0
        assert position["retMsg"] == "OK"



import base64
import hmac
import logging
import time
from contextlib import asynccontextmanager

import trio
import orjson
import trio_websocket
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from trio_websocket import open_websocket_url

from trio_bybit.exceptions import BybitWebsocketOpError


class BybitSocketManager:
    URLS = {
        "main": {
            "spot": "wss://stream.bybit.com/v5/public/spot",
            "linear": "wss://stream.bybit.com/v5/public/linear",
            "inverse": "wss://stream.bybit.com/v5/public/inverse",
            "private": "wss://stream.bybit.com/v5/private",
        },
        "test": {
            "spot": "wss://stream-testnet.bybit.com/v5/public/spot",
            "linear": "wss://stream-testnet.bybit.com/v5/public/linear",
            "inverse": "wss://stream-testnet.bybit.com/v5/public/inverse",
            "private": "wss://stream-testnet.bybit.com/v5/private",
        },
        "demo": {
            "private": "wss://stream-demo.bybit.com",
        },
    }
    ws: trio_websocket.WebSocketConnection
    cancel_scope: trio.CancelScope
    conn_id: str
    last_pong: int  # milliseconds

    def __init__(
        self,
        endpoint: str = "spot",
        api_key: str | None = None,
        api_secret: str | None = None,
        api_secret_passphrase: bytes | None = None,
        alternative_net: str = "",
        sign_style: str = "HMAC",
    ):
        self.endpoint: str = endpoint
        self.alternative_net: str = alternative_net if alternative_net else "main"
        if self.endpoint == "private" and (api_key is None or api_secret is None):
            raise ValueError("api_key and api_secret must be provided for private streams")
        self.api_key = api_key
        self.api_secret = api_secret
        if sign_style != "HMAC":
            assert api_secret is not None
            with open(api_secret, "rb") as f:
                self.api_secret = load_pem_private_key(f.read(), password=api_secret_passphrase)
        else:
            self.api_secret = api_secret
        self.sign_style = sign_style

        self.connected = trio.Condition()
        self.subscribed = set()  # topics subscribed, for re-subscription

    @asynccontextmanager
    async def connect(self):
        """
        Asynchronous context manager to establish and maintain a WebSocket connection.

        This method attempts to connect to the specified WebSocket URL and manages the connection lifecycle.
        If the connection is closed, it will automatically attempt to reconnect.

        Note: this method will not end until forced to. A trio cancel scope could help.

        Yields:
            trio_websocket.WebSocketConnection: The active WebSocket connection.

        Raises:
            ValueError: If the specified endpoint and network combination is not supported.
            RuntimeError: If the connection context manager exits unexpectedly.
        """
        try:
            url = self.URLS[self.alternative_net][self.endpoint]
        except KeyError:
            raise ValueError(f"endpoint {self.endpoint} with net {self.alternative_net} not supported")

        async def _conn(task_status=trio.TASK_STATUS_IGNORED):
            with trio.CancelScope() as scope:
                async with open_websocket_url(url) as websock:
                    self.ws = websock
                    if self.endpoint == "private":
                        await self._send_signature()
                    if self.subscribed:
                        await self.subscribe({"op": "subscribe", "args": list(self.subscribed)})
                    task_status.started(scope)
                    async with self.connected:
                        self.connected.notify_all()
                    await self.heartbeat()

        while True:
            async with trio.open_nursery() as nursery:
                self.cancel_scope = await nursery.start(_conn)
                yield self.ws

            if self.cancel_scope.cancelled_caught:  # connection closed
                logging.info("Connection closed, restarting...")
                continue  # restarting connection
            else:  # should not come here
                raise RuntimeError("Unexpected exit from websocket connection context manager.")

    async def _send_message(self, message: str | bytes):
        """
        Sends a message through the WebSocket connection.

        This method ensures that messages are sent through the WebSocket connection
        and handles the `ConnectionClosed` exception by cancelling the current scope,
        which will trigger a reconnection.

        Parameters:
            message (str | bytes): The message to be sent through the WebSocket.
        """
        try:
            await self.ws.send_message(message)
        except trio_websocket.ConnectionClosed:
            self.cancel_scope.cancel()
            async with self.connected:
                await self.connected.wait()
                await self.ws.send_message(message)

    async def _get_message(self) -> str | bytes:
        """
        Retrieve a message from the WebSocket connection.

        This method attempts to get a message from the WebSocket connection. If the connection
        is closed, it cancels the current scope to trigger a reconnection.

        Returns:
            str | bytes: The message received from the WebSocket connection.
        """
        try:
            return await self.ws.get_message()
        except trio_websocket.ConnectionClosed:
            self.cancel_scope.cancel()
            async with self.connected:
                await self.connected.wait()
                return await self.ws.get_message()

    async def heartbeat(self):
        while True:
            with trio.fail_after(5):
                await self._send_message('{"op": "ping"}')
            await trio.sleep(20)

    async def _send_signature(self):
        expires = int((time.time() + 1) * 1000)
        if self.sign_style == "HMAC":
            assert isinstance(self.api_secret, str)
            signature = str(
                hmac.new(
                    self.api_secret.encode("utf-8"), f"GET/realtime{expires}".encode("utf-8"), digestmod="sha256"
                ).hexdigest()
            )
        else:  # RSA
            assert isinstance(self.api_secret, RSAPrivateKey)
            signature = self.api_secret.sign(
                f"GET/realtime{expires}".encode("utf-8"), padding.PKCS1v15(), hashes.SHA256()
            )
            signature = base64.b64encode(signature).decode()
        await self._send_message(orjson.dumps({"op": "auth", "args": [self.api_key, expires, signature]}))
        auth_ret = orjson.loads(await self._get_message())
        if auth_ret["op"] == "auth":
            try:
                assert auth_ret["success"]
            except AssertionError:
                raise BybitWebsocketOpError(auth_ret)
            self.conn_id = auth_ret["conn_id"]

    async def subscribe(self, subscription: dict):
        """
        Subscribe or unsubscribe to a websocket stream.

        Parameters:
            subscription (dict): (un)subscription message, e.g. {"op": "subscribe", "args": ["publicTrade.BTCUSDT"]}
        """
        if subscription["op"] == "subscribe":
            self.subscribed.update(subscription["args"])
        elif subscription["op"] == "unsubscribe":
            self.subscribed.difference_update(subscription["args"])
        else:
            raise ValueError(f"op must be 'subscribe' or 'unsubscribe', but received '{subscription["op"]}'")
        await self._send_message(orjson.dumps(subscription))

    async def get_next_message(self):
        """
        Continuously retrieves messages from the WebSocket connection.

        Yields:
            dict: The message received from the WebSocket connection containing "topic" and "data".
        """
        while True:
            raw_message = await self._get_message()
            message = orjson.loads(raw_message)
            if "topic" in message and "data" in message:
                yield message
                continue
            elif "op" in message:
                if message["op"] == "pong":
                    self.last_pong = int(message["args"][0])
                elif not message.get("success"):  # probably a subscription error
                    raise BybitWebsocketOpError(raw_message)



import logging
import trio

from trio_bybit import AsyncClient
from trio_bybit.streams import BybitSocketManager


async def main():
    socket = BybitSocketManager(
        endpoint="private",
        api_key="GV0XaFwrE8Z7k0QHoy",
        api_secret="nmsmcyWGdvbxD7KDDBz4RttUjFeK7ijRNU1Q",
    )
    async with socket.connect():
        subscription = {
            "op": "subscribe",
            "args": ["order"],
        }
        await socket.subscribe(subscription)

        async for msg in socket.get_next_message():
            print(msg)


if __name__ == "__main__":

    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')


    trio.run(main)
